{# pe_df_result.html #}

{% extends "base.html" %}

{% block extrahead %}
    <style type="text/css">{{ sql_highlight_css }}</style>
{% endblock %}

{% block content %}

    {% include "query_nav.html" %}

    <h1>Patient Explorer: Data Finder results</h1>

    <h2>Results (up to {{ patients_per_page }}
        patient{{ patients_per_page|pluralize }} per page)</h2>

    {% if some_patients %}

        {% include "page_nav_results.html" with page=page %}

        {{ results_table_html|safe }}

        {% include "page_nav_results.html" with page=page %}

        <h2>Query</h2>
        {{ query_html|safe }}

    {% else %}
        <div class="warning">No patients found.</div>
    {% endif %}

    <h2>Common patient ID query</h2>
    {{ patient_id_query_html|safe }}

    {% if some_patients %}
        {% include "page_nav_results.html" with page=page %}
    {% endif %}

{% endblock %}
