/* base.css */

body {
    font-family: Arial, Helvetica, sans-serif;
    text-align: left;
    line-height: 1.2em;
}
div {
    margin: 0.5em 0;
}
em {
    color: rgb(0, 94, 184); /* NHS Blue; rgb(0, 94, 184) = #005EB8; https://www.england.nhs.uk/nhsidentity/identity-guidelines/colours/ */
    font-style: normal;
}
pre {
    margin: 0;
}
span {
    display: inline;
}
table {
    display: table;  /* with table-cell, auto-resizing goes wrong */
    border-collapse: collapse;
    vertical-align: top;
    margin: 0.25em 0;
    /* width: 100%; */
}
tr {
    display: table-row;
}
th, td {
    display: table-cell;
    vertical-align: top;
    border: 1px solid black;
    margin: 0.25em;
    padding: 0.1em 0.3em;
}
p {
    margin: 0.4em 0;
}
ul {
    margin: 0.25em 0;
}

.addresses {
    clear: both;
}
.address_to {
    float: left;
}
.address_from {
    float: right;
    text-align: right;
}
.action {
    color: rgb(0, 0, 150);
}
.collapsible {
    margin: 0;
}
.collapse_visible {}
.collapse_invisible {
    display: none;
}
.collapse_big {}
.collapse_small {
    background-color: rgb(240, 240, 255);
    max-height: 5em;
    overflow: hidden;
}

.consent {
    background-color: rgb(230, 230, 230);
    border: 0.5mm solid black;
    padding: 0.5em;
    margin: 0.25em 0;
}
.consent tr, .consent th, .consent td {
    border-style: none;
    font-size: 0.9em;
}
.consent .persontype {
    vertical-align: bottom;
    font-size: 0.8em;
    line-height: 1.2em;
}
.consent .label {
    text-align: right;
    vertical-align: bottom;
    font-size: 0.8em;
    line-height: 1.2em;
}
.consent .footnote {
    font-size: 0.7em;
    line-height: 1.2em;
}
.deadcell {
    background-color: rgb(200, 200, 200);
}
.error {
    color: rgb(255, 0, 0);
}

/*noinspection CssUnusedSymbol*/
.errorlist {  /* Django forms */
    color: rgb(255, 0, 0);
}
.expandcollapsewrapper {
    display: table;
    margin: 0;
}
.expandcollapse {
    display: table-cell;
    margin: 0;
}
.expandcollapse_span {
    display: inline;
    margin: 0;
}
.formtable, .formtable th, .formtable tr, .formtable td {
    border: none;
}
.formtitle {
    font-size: 1.6em;
    line-height: 1.2em;
    color: rgb(0, 114, 198);
}

.hangingindent {
    padding-left: 1.5em;
    text-indent: -1.5em;
}
.highlight {
    background-color: rgb(255, 255, 224);
}

/* yellow and red are typically used by browsers for all-instances and
   instance-at-the-cursor with Ctrl-F */
/*noinspection CssUnusedSymbol*/
.highlight0 { background-color: #0F0; }
/*noinspection CssUnusedSymbol*/
.highlight1 { background-color: #88F; }
/*noinspection CssUnusedSymbol*/
.highlight2 { background-color: #F0F; }
/* If altering: change also N_CSS_HIGHLIGHT_CLASSES in the source */

.important {
    color: rgb(150, 0, 0);
}
.indent {
    margin-left: 3em;
}
.letterbody {
    clear: both;
}
.letterdate {
    text-align: right;
}
.logo {
    text-align: right;
}
.office {
    background-color: rgb(200, 200, 200);
    padding: 0.5em;
}

.options {
    border: none;
    width: 100%;
    vertical-align: top;
}
table.options, .options table {
    border-collapse: separate;
    border-spacing: 0 0.5em;  /* horizontal, vertical */
}
.options tr, .options th, .options td {
    border: none;
}

.patient {
    background-color: rgb(127, 255, 127);
}
.patient_withdrawal {
    background-color: rgb(255, 127, 127);
}
.patient, .patient_withdrawal {
    /* width: 100%; */
}
.patient_details {
    margin-left: 3em;
    font-style: italic;
}
.patientprefs {
    border: 0.5mm solid black;
}
.patientprefs tr {
    border-style: none;
}
.patientprefs th {
    border-style: solid none;
}
.patientprefs td {
    border-style: none;
}
.rjust {
    text-align: right;
}
.signature {
    width: 100%;
    border: none;
    padding: 0.5em;
    table-layout: fixed;
}
.signature tr, .signature th, .signature td {
    border: 0.5mm none black;
}
.signature_label {
    text-align: right;
    vertical-align: bottom;
}
.signature_box, .consent .signature_box, .signature .signature_box {
    /* Increase CSS specificity to have this override ".signature td" etc. */
    padding-top: 2.5em;
    border-bottom: 0.5mm solid black;
}
.email_label {
    text-align: left;
    vertical-align: bottom;
}
.email_box, .options .email_box {
    /* Increase CSS specificity to have this override ".options td" etc. */
    padding-top: 0.5em;
    border-bottom: 0.5mm solid black;
}

.study {
    background-color: rgb(200, 200, 255);
}
.titlecolour {
    color: rgb(0, 114, 198);
}
.underline {
    text-decoration: underline;
}
.queryresult {
    /* font-family: Georgia, "Times New Roman", Times, serif; */
    font-family: "Times New Roman", Times, serif;
}
.ditto {
    font-style: italic;
    font-weight: bold;
    text-align: center;
    background-color: rgb(230, 230, 230);
}
.stripy_odd {
}
.stripy_even {
    background-color: rgb(245, 255, 230);
}
.sql {
    background-color: rgb(230, 230, 255);
}
.tickbox {
    font-family: 'Consolas', 'Monaco', 'Bitstream Vera Sans Mono', monospace;
    text-align: right;
}
.warning {
    font-weight: bold;
    color: rgb(255, 0, 0);
}
.code {
    font-family: monospace;
}

.page_break_before {
    page-break-before: always;
}
